# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from ogr.services.forgejo.comments import (
    ForgejoComment,
    ForgejoIssueComment,
    ForgejoPRComment,
)
from ogr.services.forgejo.issue import ForgejoIssue
from ogr.services.forgejo.project import ForgejoProject
from ogr.services.forgejo.pull_request import ForgejoPullRequest
from ogr.services.forgejo.release import ForgejoRelease
from ogr.services.forgejo.service import ForgejoService
from ogr.services.forgejo.user import ForgejoUser

__all__ = [
    ForgejoPullRequest.__name__,
    ForgejoComment.__name__,
    ForgejoIssueComment.__name__,
    ForgejoPRComment.__name__,
    ForgejoIssue.__name__,
    ForgejoProject.__name__,
    ForgejoRelease.__name__,
    ForgejoService.__name__,
    ForgejoUser.__name__,
]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT


from requre.helpers import record_httpx
from requre.online_replacing import record_requests_for_all_methods

from tests.integration.forgejo.base import ForgejoTests


@record_httpx()
@record_requests_for_all_methods()
class Comments(ForgejoTests):
    def test_pr_comments(self):
        pr_comments = list(self.project.get_pr(209).get_comments())
        assert pr_comments
        assert len(pr_comments) == 3

        assert pr_comments[0].body.startswith("LGTM")
        assert pr_comments[1].body.startswith("second comment")

    def test_pr_comments_reversed(self):
        pr_comments = list(self.project.get_pr(209).get_comments(reverse=True))
        assert pr_comments
        assert len(pr_comments) == 3

        assert pr_comments[0].body.startswith("LGTM, nicely done")

    def test_pr_comments_filter(self):
        pr_comments = list(
            self.project.get_pr(209).get_comments(filter_regex="comment"),
        )
        assert pr_comments
        assert len(pr_comments) == 1
        assert pr_comments[0].body.startswith("second")

        pr_comments = list(
            self.project.get_pr(209).get_comments(
                filter_regex="nicely ([a-z]*)",
            ),
        )
        assert pr_comments
        assert len(pr_comments) == 1
        assert pr_comments[0].body.endswith("done")

    def test_pr_comments_search(self):
        comment_match = self.project.get_pr(209).search(filter_regex="LGTM")
        assert comment_match
        assert comment_match[0] == "LGTM"

        comment_match = self.project.get_pr(209).search(
            filter_regex="LGTM, nicely ([a-z]*)",
        )
        assert comment_match
        assert comment_match[0] == "LGTM, nicely done"

    def test_issue_comments(self):
        comments = list(self.project.get_issue(244).get_comments())
        assert len(comments) == 3
        assert comments[0].body.startswith("/packit")

    def test_issue_comments_reversed(self):
        comments = list(self.project.get_issue(244).get_comments(reverse=True))
        assert len(comments) == 3
        assert comments[0].body.startswith("This ")

    def test_issue_comments_regex(self):
        comments = list(
            self.project.get_issue(244).get_comments(
                filter_regex=".*random *",
            ),
        )
        assert len(comments) == 2
        assert "other" in comments[0].body

    def test_issue_comments_regex_reversed(self):
        comments = list(
            self.project.get_issue(244).get_comments(
                reverse=True,
                filter_regex=".*random *",
            ),
        )
        assert len(comments) == 2
        assert "indeed" in comments[0].body

    def test_pr_comments_author_regex(self):
        comments = list(
            self.project.get_pr(209).get_comments(
                filter_regex=".*nic*",
                author="packit-validator",
            ),
        )
        assert len(comments) == 1
        assert comments[0].author == "packit-validator"
        assert "nicely" in comments[0].body

    def test_pr_comments_author(self):
        comments = list(
            self.project.get_pr(209).get_comments(author="packit-validator"),
        )
        assert len(comments) == 3
        assert comments[0].author == "packit-validator"
        assert comments[0].body.endswith("TM")

    def test_issue_comments_author_regex(self):
        comments = list(
            self.project.get_issue(244).get_comments(
                filter_regex=".*test-*",
                author="packit-validator",
            ),
        )
        assert len(comments) == 1
        assert comments[0].author == "packit-validator"
        assert comments[0].body.startswith("/packit")

    def test_issue_comments_author(self):
        comments = list(
            self.project.get_issue(244).get_comments(author="packit-validator"),
        )

        assert len(comments) == 3
        assert comments[0].author == "packit-validator"
        assert comments[0].body.startswith("/packit")
        assert comments[2].body.endswith("indeed")

    def test_issue_comments_updates(self):
        comments = list(
            self.project.get_issue(244).get_comments(
                filter_regex="test-comment",
            ),
        )
        assert len(comments) == 1
        before_comment = comments[0].body
        before_edited = comments[0].edited

        comments[0].body = "here comes the update"
        assert comments[0].body == "here comes the update"

        # using >= because the time difference is so small that the datetime values are the same
        assert comments[0].edited >= before_edited

        comments[0].body = before_comment
        assert comments[0].body == before_comment

    def test_pr_comments_updates(self):
        comments = list(
            self.project.get_pr(209).get_comments(
                filter_regex="nd comment",
            ),
        )
        assert len(comments) == 1
        before_comment = comments[0].body
        before_edited = comments[0].edited

        comments[0].body = "this wont hurt a bit"
        assert comments[0].body == "this wont hurt a bit"

        # using >= because the time difference is so small that the datetime values are the same
        assert comments[0].edited >= before_edited

        comments[0].body = before_comment
        assert comments[0].body == before_comment

    def test_pr_react_to_comment_and_delete(self):
        pr = self.project.get_pr(199)
        pr_comment = pr.comment("React to this comment")

        reaction = pr_comment.add_reaction("+1")
        assert len(pr_comment.get_reactions()) == 1

        reaction.delete()
        assert not reaction._raw_reaction
        assert len(pr_comment.get_reactions()) == 0

    def test_issue_react_to_comment_and_delete(self):
        issue = self.project.get_issue(245)
        issue_comment = issue.comment("This is a nice comment to react to")

        reaction = issue_comment.add_reaction("confused")
        assert len(issue_comment.get_reactions()) == 1

        reaction.delete()
        assert not reaction._raw_reaction
        assert len(issue_comment.get_reactions()) == 0

    def test_pr_get_reactions(self):
        pr = self.project.get_pr(199)
        pr_comment = pr.comment("Such nice weather")

        pr_comment.add_reaction("+1")
        pr_comment.add_reaction("-1")
        pr_comment.add_reaction("confused")
        assert len(pr_comment.get_reactions()) == 3

    def test_issue_get_reactions(self):
        issue = self.project.get_issue(245)
        issue_comment = issue.comment("Nice weather indeed")

        issue_comment.add_reaction("+1")
        issue_comment.add_reaction("-1")
        issue_comment.add_reaction("confused")
        assert len(issue_comment.get_reactions()) == 3



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import functools
import logging
import re
import time
from collections.abc import Iterable
from re import Match
from typing import Any, Callable, Optional, Union

from urllib3.util import Retry

try:
    # If urllib3~=2.0 is installed
    from urllib3 import BaseHTTPResponse
except ImportError:
    # If urllib3~=1.0 is installed
    from urllib3 import HTTPResponse as BaseHTTPResponse


from ogr.abstract import AnyComment, Comment

logger = logging.getLogger(__name__)


class CustomRetry(Retry):
    """
    Custom Retry class that includes 403 in RETRY_AFTER_STATUS_CODES
    so that Retry-After headers are respected for 403 errors.

    Also handles GitHub rate limit headers (X-RateLimit-Reset) when
    Retry-After is not present.
    """

    # Include 403 in the list of status codes that respect Retry-After header
    RETRY_AFTER_STATUS_CODES = frozenset([413, 429, 503, 403])

    def get_ratelimit_reset(self, response: BaseHTTPResponse) -> Optional[float]:
        """
        Get retry wait time from X-RateLimit-Reset header.

        Rate limit reset header (Unix timestamp) which is converted
        to seconds to wait, compatible with Retry-After format.

        Args:
            response: HTTP response object that may contain X-RateLimit-Reset header.

        Returns:
            Number of seconds to wait before retrying, or None if header is not present
            or cannot be parsed.
        """
        # Only check X-RateLimit-Reset for rate limit responses
        if (  # noqa: SIM102 This is more readable than a single if statement
            response.status
            in (
                403,
                429,
            )
        ):
            # urllib3 HTTPHeaderDict does a case-insensitive lookup
            # https://github.com/urllib3/urllib3/blob/83f8643ffb5b7f197457379148e2fa118ab0fcdc/src/urllib3/_collections.py#L215-L217
            if rate_limit_reset := response.headers.get(
                "X-RateLimit-Reset",
            ):
                try:
                    reset_timestamp = float(rate_limit_reset)
                except ValueError:
                    logger.error(
                        f"Could not parse X-RateLimit-Reset header '{rate_limit_reset}'",
                    )
                    return None
                else:
                    return max(0.0, reset_timestamp - time.time())
        return None

    def sleep_for_retry(self, response: BaseHTTPResponse) -> bool:
        """
        Override to handle X-RateLimit-Reset header in addition to Retry-After.

        Choose between Retry-After and X-RateLimit-Reset header.
        If both are present, choose the longer wait time.

        Args:
            response: HTTP response object that may contain Retry-After or X-RateLimit-Reset header.

        Returns:
            True if the wait time is greater than 0, False otherwise.
        """
        retry_after = self.get_retry_after(response)
        rate_limit_reset = self.get_ratelimit_reset(response)

        if not retry_after and not rate_limit_reset:
            return False

        wait_time, header = max(
            (
                (retry_after or 0, "Retry-After"),
                (rate_limit_reset or 0, "X-RateLimit-Reset"),
            ),
            key=lambda x: x[0],
        )
        logger.error(
            f"Rate limit hit (status {response.status}). "
            f"Waiting {wait_time}s until reset ({header} header)",
        )
        time.sleep(wait_time)
        return True


def filter_comments(
    comments: Union[list[AnyComment], Iterable[AnyComment]],
    filter_regex: Optional[str] = None,
    author: Optional[str] = None,
) -> Union[list[AnyComment], Iterable[AnyComment]]:
    """
    Filters comments from the given list.

    Args:
        comments: List of comments to be filtered.
        filter_regex: Regex to be used for filtering body of the
            comments.

            Defaults to `None`, which means no filtering by regex.
        author: Login of the author of the comments.

            Defaults to `None`, which means no filtering by author.

    Returns:
        List of comments that satisfy requested criteria.
    """
    if filter_regex or author:
        pattern = None
        if filter_regex:
            pattern = re.compile(filter_regex)

        comments = list(
            filter(
                lambda comment: (not pattern or bool(pattern.search(comment.body)))
                and (not author or comment.author == author),
                comments,
            ),
        )
    return comments


def search_in_comments(
    comments: Iterable[Union[str, Comment]],
    filter_regex: str,
) -> Optional[Match[str]]:
    """
    Find match in pull request description or comments.

    Args:
        comments: List of comments or bodies of comments
            to be searched through.
        filter_regex: Regex to be used for filtering with `re.search`.

    Returns:
        Match that has been found, `None` otherwise.
    """
    pattern = re.compile(filter_regex)
    for comment in comments:
        if isinstance(comment, Comment):
            comment = comment.body
        re_search = pattern.search(comment)
        if re_search:
            return re_search
    return None


class RequestResponse:
    """
    Class that holds response for Pagure requests.

    Attributes:
        status_code (int): Status code of the response.
        ok (bool): `True` if successful, `False` otherwise.
        content (bytes): Content of the response.
        json_content (Optional[Dict[Any, Any]]): JSON content of the response.
    """

    def __init__(
        self,
        status_code: int,
        ok: bool,
        content: bytes,
        json: Optional[dict[Any, Any]] = None,
        reason: Optional[str] = None,
        headers: Optional[list[tuple[Any, Any]]] = None,
        links: Optional[list[str]] = None,
        exception: Optional[dict[Any, Any]] = None,
    ) -> None:
        self.status_code = status_code
        self.ok = ok
        self.content = content
        self.json_content = json
        self.reason = reason
        self.headers = dict(headers) if headers else None
        self.links = links
        self.exception = exception

    def __str__(self) -> str:
        return (
            f"RequestResponse("
            f"status_code={self.status_code}, "
            f"ok={self.ok}, "
            f"content={self.content.decode()}, "
            f"json={self.json_content}, "
            f"reason={self.reason}, "
            f"headers={self.headers}, "
            f"links={self.links}, "
            f"exception={self.exception})"
        )

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, RequestResponse):
            return False
        return (
            self.status_code == o.status_code
            and self.ok == o.ok
            and self.content == o.content
            and self.json_content == o.json_content
            and self.reason == o.reason
            and self.headers == o.headers
            and self.links == o.links
            and self.exception == o.exception
        )

    def to_json_format(self) -> dict[str, Any]:
        """
        Returns:
            Response in a JSON format.
        """
        output = {
            "status_code": self.status_code,
            "ok": self.ok,
            "content": self.content,
        }
        if self.json_content:
            output["json"] = self.json_content
        if self.reason:
            output["reason"] = self.reason
        if self.headers:
            output["headers"] = self.headers
        if self.links:
            output["links"] = self.links
        if self.exception:
            output["exception"] = self.exception
        return output

    def json(self) -> Optional[dict[Any, Any]]:
        """
        Returns:
            JSON content of the response.
        """
        return self.json_content


def filter_paths(paths: Iterable[str], filter_regex: str) -> Iterable[str]:
    """
    Filters paths from the given list.

    Args:
        paths: List of paths to be filtered, in a form of an iterable.
        filter_regex: Regex to be used for filtering paths.

    Returns:
        List of path that satisfy regex, in a from of an iterable.
    """
    pattern = re.compile(filter_regex)
    return (path for path in paths if not pattern or bool(pattern.search(path)))


def indirect(specialized_function: Callable) -> Any:
    """
    Decorator to wrap methods on `GitProject`s that call specialized classes.

    Args:
        specialized_function: Static method of the specialized class
            that takes as first argument the `GitProject` itself.

    Returns:
        Decorator that calls `specialized_function` once called.
    """

    def indirect_caller(func):
        @functools.wraps(func)
        def indirectly_called(self, *args, **kwargs):
            return specialized_function(self, *args, **kwargs)

        return indirectly_called

    return indirect_caller


def create_retry_config(max_retries: Union[int, Retry]) -> Retry:
    """
    Create a retry configuration for the given max retries.
    Apply suggestions from https://docs.github.com/en/rest/using-the-rest-api/troubleshooting-the-rest-api?apiVersion=2022-11-28

    Args:
        max_retries: Maximum number of retries.

    Returns:
        Retry configuration.
    """

    if isinstance(max_retries, Retry):
        return max_retries
    return CustomRetry(
        total=int(max_retries),
        # Retry mechanism active for these HTTP methods:
        allowed_methods=["DELETE", "GET", "PATCH", "POST", "PUT"],
        # Only retry on following HTTP status codes
        status_forcelist=[500, 503, 403, 401, 429],
        # This helps when hitting rate limits or temporary server issues
        # Exponential backoff: wait 30s, 60s, 120s between retries
        backoff_factor=30,
        # Respect Retry-After header for status codes in RETRY_AFTER_STATUS_CODES
        # (413, 429, 503, 403). CustomRetry includes 403 so Retry-After will be
        # respected for 403 errors when present.
        respect_retry_after_header=True,
        raise_on_status=True,
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
import itertools
from collections.abc import Iterable
from datetime import datetime
from functools import partial
from typing import Optional, Union

from pyforgejo import NotFoundError
from pyforgejo.core.api_error import ApiError
from pyforgejo.types import User as PyforgejoUser
from pyforgejo.types.issue import Issue as _issue

from ogr.abstract import Issue, IssueComment, IssueLabel, IssueStatus
from ogr.exceptions import (
    ForgejoAPIException,
    IssueTrackerDisabled,
    OperationNotSupported,
)
from ogr.services import forgejo
from ogr.services.base import BaseIssue
from ogr.services.forgejo.comments import ForgejoIssueComment
from ogr.services.forgejo.label import ForgejoIssueLabel
from ogr.services.forgejo.utils import paginate


class ForgejoIssue(BaseIssue):
    project: "forgejo.ForgejoProject"

    def __init__(self, raw_issue: _issue, project: "forgejo.ForgejoProject"):
        if raw_issue.pull_request:
            raise ForgejoAPIException(
                f"Requested issue #{raw_issue.number} is a pull request",
            )

        super().__init__(raw_issue, project)

    @property
    def api(self):
        """Returns the issue API client from pyforgejo."""
        return self.project.service.api.issue

    def partial_api(self, method, /, *args, **kwargs):
        """Returns a partial API call for ForgejoIssue.


        Injects owner and repo parameters for the calls to issue API endpoints.

        Args:
            method: Specific method on the Pyforgejo API that is to be wrapped.
            *args: Positional arguments that get injected into every call.
            **kwargs: Keyword-arguments that get injected into every call.

        Returns:
            Callable with pre-injected parameters.
        """
        params = {"owner": self.project.namespace, "repo": self.project.repo}
        return partial(method, *args, **kwargs, **params)

    def __update_info(self) -> None:
        """Refresh the local issue object with the latest data from the server."""
        self._raw_issue = self.partial_api(self.api.get_issue)(index=self.id)

    @property
    def title(self) -> str:
        return self._raw_issue.title

    @title.setter
    def title(self, new_title: str) -> None:
        self._raw_issue = self.partial_api(self.api.edit_issue)(
            title=new_title,
            index=self.id,
        )

    @property
    def id(self) -> int:
        return self._raw_issue.number

    @property
    def url(self) -> str:
        return self._raw_issue.url

    @property
    def description(self) -> str:
        return self._raw_issue.body

    @description.setter
    def description(self, new_description: str):
        self._raw_issue = self.partial_api(self.api.edit_issue)(
            body=new_description,
            index=self.id,
        )

    @property
    def author(self) -> str:
        return self._raw_issue.user.login

    @property
    def created(self) -> datetime:
        return self._raw_issue.created_at

    @property
    def status(self) -> IssueStatus:
        return IssueStatus[self._raw_issue.state]

    @property
    def assignees(self) -> list[PyforgejoUser]:
        return self._raw_issue.assignees or []

    @property
    def labels(self) -> list[IssueLabel]:
        return [
            ForgejoIssueLabel(raw_label, self) for raw_label in self._raw_issue.labels
        ]

    def __str__(self) -> str:
        return "Forgejo" + super().__str__()

    @staticmethod
    def create(
        project: "forgejo.ForgejoProject",
        title: str,
        body: str,
        private: Optional[bool] = None,
        labels: Optional[list[str]] = None,
        assignees: Optional[list[str]] = None,
    ) -> "Issue":
        if private:
            raise OperationNotSupported("Private issues are not supported by Forgejo")
        if not project.has_issues:
            raise IssueTrackerDisabled()

        # The API requires ids of labels in the create_issue method
        # which would lead to having to retrieve existing labels and
        # needing to find the ids of those we need to add to the issue;
        # A separate API call would also need to be made to create each
        # label that does not yet exist, potentially leading to many
        # API calls and unclear code, so labels are instead added seprately
        # below after creating a new issue without labels
        issue = project.service.api.issue.create_issue(
            owner=project.namespace,
            repo=project.repo,
            title=title,
            body=body,
            labels=[],
            assignees=assignees,
        )

        forgejo_issue = ForgejoIssue(issue, project)

        if labels:
            forgejo_issue.add_label(*labels)

        return forgejo_issue

    @staticmethod
    def get(project: "forgejo.ForgejoProject", issue_id: int) -> "Issue":
        if not project.has_issues:
            raise IssueTrackerDisabled()

        try:
            issue = project.service.api.issue.get_issue(
                owner=project.namespace,
                repo=project.repo,
                index=issue_id,
            )

        except NotFoundError as ex:
            raise ForgejoAPIException(f"Issue {issue_id} not found") from ex
        return ForgejoIssue(issue, project)

    @staticmethod
    def get_list(
        project: "forgejo.ForgejoProject",
        status: IssueStatus = IssueStatus.open,
        author: Optional[str] = None,
        assignee: Optional[str] = None,
        labels: Optional[list[str]] = None,
    ) -> list["Issue"]:
        if not project.has_issues:
            raise IssueTrackerDisabled()

        parameters: dict[str, Union[str, list[str], bool]] = {
            "state": status.name,
            "type": "issues",
        }
        if author:
            parameters["created_by"] = author
        if assignee:
            parameters["assigned_by"] = assignee
        if labels:
            parameters["labels"] = labels

        try:
            return [
                ForgejoIssue(issue, project)
                for issue in paginate(
                    project.service.api.issue.list_issues,
                    owner=project.namespace,
                    repo=project.repo,
                    **parameters,
                )
            ]
        except NotFoundError as ex:
            raise ForgejoAPIException("Failed to list issues") from ex

    def comment(self, body: str) -> IssueComment:
        comment = self.partial_api(self.api.create_comment)(
            body=body,
            index=self.id,
        )
        return ForgejoIssueComment(parent=self, raw_comment=comment)

    def close(self) -> "Issue":
        self._raw_issue = self.partial_api(self.api.edit_issue)(
            state="closed",
            index=self.id,
        )

        return self

    def _get_all_comments(self, reverse: bool = False) -> Iterable[IssueComment]:
        comments = self.partial_api(self.api.get_comments)(
            index=self.id,
        )

        if reverse:
            comments = list(reversed(comments))

        return (
            ForgejoIssueComment(parent=self, raw_comment=raw_comment)
            for raw_comment in comments
        )

    def get_comment(self, comment_id: int) -> ForgejoIssueComment:
        return ForgejoIssueComment(
            self.partial_api(self.api.get_comment)(id=comment_id),
            parent=self,
        )

    def add_assignee(self, *assignees: str) -> None:
        current_assignees = [assignee.login for assignee in self.assignees]
        updated_assignees = set(itertools.chain(current_assignees, assignees))

        try:
            self._raw_issue = self.partial_api(self.api.edit_issue)(
                assignees=updated_assignees,
                index=self.id,
            )
        except ApiError as ex:
            raise ForgejoAPIException(
                "Failed to assign issue, unknown user",
            ) from ex

    def add_label(self, *labels: str) -> None:
        self.partial_api(self.api.add_label)(labels=labels, index=self.id)
        self.__update_info()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from requre.helpers import record_httpx
from requre.online_replacing import record_requests_for_all_methods

from ogr.abstract import IssueStatus
from ogr.exceptions import (
    ForgejoAPIException,
    GitForgeInternalError,
    IssueTrackerDisabled,
)
from tests.integration.forgejo.base import ForgejoTests


@record_httpx()
@record_requests_for_all_methods()
class Issues(ForgejoTests):

    random_str = "abcdedfgh"
    title = "Issue title"
    description = "Issue body"

    def test_get_issue_list(self):
        issue_list = self.project.get_issue_list()
        assert issue_list
        assert len(issue_list) >= 1

    def test_get_issue_list_author(self):
        issue_list = self.project.get_issue_list(
            status=IssueStatus.all,
            author="packit-validator",
        )
        assert issue_list
        assert len(issue_list) >= 3

    def test_get_issue_list_nonexisting_author(self):
        with pytest.raises(ForgejoAPIException):
            self.project.get_issue_list(
                status=IssueStatus.all,
                author="xyzidontexist",
            )

    def test_get_issue_list_assignee(self):
        issue_list = self.project.get_issue_list(
            status=IssueStatus.all,
            assignee="packit-validator",
        )
        assert issue_list
        assert len(issue_list) >= 1

    def test_issue_info(self):
        title = "test Title ABC"
        description = "This is a test issue for checking title prefix"
        issue = self.project.create_issue(title=title, body=description)
        fetched = self.project.get_issue(issue.id)
        assert fetched.title.startswith("test Title")

    def test_create_issue_with_assignee(self):
        issue_title = "This is a example issue"
        issue_desc = "Description for issue"
        assign = ["packit-validator"]
        project = self.project
        issue = project.create_issue(
            title=issue_title,
            body=issue_desc,
            assignees=assign,
        )
        assert issue.title == issue_title
        assert issue.assignees[0].login == assign[0]
        assert issue.description == issue_desc

    def test_create_issue_with_labels(self):
        issue_title = "A super duper issue"
        issue_desc = "Description for issue"

        # labels need to be sorted alphabetically as that is the order
        # in which they are returned by the API
        labels = ["b", "label2"]

        issue = self.project.create_issue(
            title=issue_title,
            body=issue_desc,
            labels=labels,
        )
        assert issue.title == issue_title
        assert issue.description == issue_desc
        assert issue.labels

        for issue_label, label in zip(issue.labels, labels):
            assert issue_label.name == label

    def test_close_issue(self):
        issue = self.project.create_issue(
            title=f"Close issue {self.random_str}",
            body=f"Description for issue for closing {self.random_str}",
        )
        issue_for_closing = self.project.get_issue(issue_id=issue.id)
        assert issue_for_closing.status == IssueStatus.open

        issue.close()
        assert issue.status == IssueStatus.closed

    def test_issue_assignees(self):
        project = self.project
        assignees = project.get_issue(223).assignees
        assert len(assignees) == 1
        assert assignees[0].login == "packit-validator"

    def test_issue_updates(self):
        issue = self.project.get_issue(issue_id=224)
        old_comments = list(issue.get_comments())
        issue.comment("test comment")
        new_comments = list(issue.get_comments())
        assert len(new_comments) > len(old_comments)

    def test_issue_labels(self):
        """
        Remove the labels from this issue before regenerating the response files:
        https://v10.next.forgejo.org/packit-validator/ogr-tests/issues/224
        """
        issue = self.project.get_issue(224)
        labels = issue.labels

        assert not labels
        issue.add_label("test_lb1", "test_lb2")
        labels = self.project.get_issue(224).labels
        assert labels
        assert next(iter(labels)).name == "test_lb1"

    def test_issue_add_assignee(self):
        """
        Remove the assignees from this issue before regenerating the response files:
        https://v10.next.forgejo.org/packit-validator/ogr-tests/issues/245
        """
        issue = self.project.get_issue(245)
        print(self.service.user.get_username())
        assignees = issue.assignees

        assert not assignees
        issue.add_assignee("packit-validator")
        assignees = self.project.get_issue(245).assignees
        assert len(assignees) == 1
        assert assignees[0].login == "packit-validator"

    def test_issue_add_assignee_without_redundant_api_call(self):
        """
        Remove the assignees from this issue before regenerating the response files:
        https://v10.next.forgejo.org/packit-validator/ogr-tests/issues/245
        """
        issue = self.project.get_issue(245)
        print(self.service.user.get_username())
        assignees = issue.assignees

        assert not assignees
        issue.add_assignee("packit-validator")
        assignees = issue.assignees
        assert len(assignees) == 1
        assert assignees[0].login == "packit-validator"

    def test_issue_no_such_assignee(self):
        issue = self.project.get_issue(245)

        # __check_for_internal_failure replaces ForgejoAPIException with GitForgeInternalError
        with pytest.raises(GitForgeInternalError):
            issue.add_assignee("nonexistentuser")

    def test_list_contains_only_issues(self):
        issue_list_all = self.project.get_issue_list(status=IssueStatus.all)
        issue_ids = [issue.id for issue in issue_list_all]

        pr_ids = [143, 170, 194, 195, 198, 199, 204, 209, 220, 221]
        for id in pr_ids:
            assert id not in issue_ids

    def test_issue_doesnt_exist(self):
        with pytest.raises(ForgejoAPIException):
            self.project.get_issue(10**20)

    def test_functions_fail_for_pr(self):
        with pytest.raises(ForgejoAPIException):
            self.project.get_issue(221)
        with pytest.raises(ForgejoAPIException):
            self.project.get_issue(221).comment(body="should fail")
        with pytest.raises(ForgejoAPIException):
            self.project.get_issue(221).close()
        with pytest.raises(ForgejoAPIException):
            _ = self.project.get_issue(221).labels
        with pytest.raises(ForgejoAPIException):
            self.project.get_issue(221).add_label("should fail")

    def test_setters(self):
        issue = self.project.get_issue(issue_id=245)

        old_title = issue.title
        issue.title = "test title"
        assert issue.title != old_title
        assert issue.title == "test title"

        issue.title = old_title
        assert issue.title == old_title

        old_description = issue.description
        issue.description = "test description"
        assert issue.description != old_description
        assert issue.description == "test description"

        issue.description = old_description
        assert issue.description == old_description

    def test_get_comment(self):
        comment = self.project.get_issue(244).get_comment(3174)
        assert comment.body == "/packit test-comment"

    def test_get_with_disabled_issues(self):
        forks = self.project.get_forks()
        fork = next(iter(forks))
        assert fork.namespace == "mfocko"

        with pytest.raises(IssueTrackerDisabled):
            fork.get_issue(issue_id=245)

    def test_get_with_id_of_pr(self):
        pr_id = 221

        # should raise an error when retrieving
        # an issue, but providing the id of a PR instead
        with pytest.raises(ForgejoAPIException):
            self.project.get_issue(pr_id)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
import logging

from pydantic import ValidationError
from pyforgejo.core.api_error import ApiError
from pyforgejo.types import Comment as _ForgejoComment
from pyforgejo.types.reaction import Reaction as _ForgejoReaction

from ogr.abstract import Comment, IssueComment, PRComment, Reaction
from ogr.services import forgejo

logger = logging.getLogger(__name__)


class ForgejoReaction(Reaction):
    def __init__(
        self,
        raw_reaction: _ForgejoReaction,
        parent: "forgejo.ForgejoComment",
    ) -> None:
        super().__init__(raw_reaction)
        self._parent = parent

    def __str__(self) -> str:
        return "Forgejo" + super().__str__()

    def delete(self) -> None:
        # self._raw_reaction is set to None after deleting
        self._raw_reaction = self._parent._client.issue.delete_comment_reaction(
            owner=self._parent._parent.project.namespace,
            repo=self._parent._parent.project.repo,
            id=self._parent._id,
        )


class ForgejoComment(Comment):
    def _from_raw_comment(self, raw_comment: _ForgejoComment) -> None:
        self._raw_comment = raw_comment
        self._id = raw_comment.id
        self._author = raw_comment.user.login
        self._created = raw_comment.created_at
        self._edited = raw_comment.updated_at

    @property
    def body(self) -> str:
        return self._raw_comment.body

    @body.setter
    def body(self, new_body: str) -> None:
        self._raw_comment = self._client.issue.edit_comment(
            owner=self._parent.project.namespace,
            repo=self._parent.project.repo,
            id=self._id,
            body=new_body,
        )

    @property
    def edited(self) -> datetime.datetime:
        return self._edited

    @property
    def _client(self):
        return self._parent.project.service.api

    def get_reactions(self) -> list[Reaction]:
        try:
            reactions = self._client.issue.get_comment_reactions(
                owner=self._parent.project.namespace,
                repo=self._parent.project.repo,
                id=self._id,
            )

        # in case no reactions exist:
        # re-recording causes pyforgejo to raise pydantic's ValidationError
        # re-running tests causes pyforgejo's ApiError to be raised instead
        # both need to be caught
        except (ValidationError, ApiError):
            return []

        return [
            ForgejoReaction(raw_reaction=reaction, parent=self)
            for reaction in reactions
        ]

    def add_reaction(self, reaction: str) -> Reaction:
        raw_reaction = self._client.issue.post_comment_reaction(
            owner=self._parent.project.namespace,
            repo=self._parent.project.repo,
            id=self._id,
            content=reaction,
        )
        return ForgejoReaction(raw_reaction=raw_reaction, parent=self)


class ForgejoIssueComment(ForgejoComment, IssueComment):
    def __str__(self) -> str:
        return "Forgejo" + super().__str__()


class ForgejoPRComment(ForgejoComment, PRComment):
    def __str__(self) -> str:
        return "Forgejo" + super().__str__()
